from pympler import asizeof

def get_obj_size(obj)->str:
    """return the size of given object

    Args:
        obj (object): any Python object

    Returns:
        str: size of the object
    """
    obj_size_byte = asizeof.asizeof(obj)

    if (obj_size_byte<1000):
        return f'{obj_size_byte:.3f} bytes'
    elif (obj_size_byte<1000000) and (obj_size_byte>=1000):
        return f'{obj_size_byte/1000:.2f} KB'
    elif (obj_size_byte<1000000000) and (obj_size_byte>=1000000):
        return f'{obj_size_byte/1000000:.2f} MB'
    elif (obj_size_byte>=1000000000):
        return f'{obj_size_byte/1000000000:.2f} GB'
    else:
        return f'{obj_size_byte/1000000000000:.2f} TB'




def dtype_detect(var) -> str:
    pass
    dtype=''
    return dtype


# def dtype_detect(var) -> Type:
#     if dtype == str:
#         dtype = ftype_detect(var)
#     else:
#         dtype=type(var)
#     return dtype


import re
import os

def get_projects(folder:str, suffix : str='_st')->dict:
    """scan streamlit Apps (folder with suffix under root directory
    Args:
        folder (str): path of root diretory
    Returns:
        dict: pairs of App name and its path; App name is folder name without suffix
    """
    return {
        re.split(suffix,item)[0]:os.path.join(folder, item) 
        for item in sorted(os.listdir(folder))
        if item.endswith(suffix)
        }

def get_versions(prj_path:str, suffix:str='_app'):
    return {
        re.split(suffix,item)[0]:os.path.join(prj_path, item) 
        for item in sorted(os.listdir(prj_path))
        if item.endswith(f'{suffix}.py')
        }


'''
This script defines several classes that extend the functionality of Python's built-in dict type to allow attribute-style access and provide additional methods for handling data, particularly in the context of data transformation and storage. The Feed class serves as a container for data, with the option to have a hierarchical structure for managing multiple groups of data sets. The DataSet class and its nested classes provide methods for managing and transforming data within these sets. The JSE_4df class is a custom JSON encoder that handles the serialization of pandas DataFrame objects.
'''

from typing import (
    Dict,
    List,
    Literal,
    Iterable
)

import logging
import copy
import keyword
import datetime
import json
from collections import UserDict

from xmlrpc.client import boolean
from pandas.core.frame import DataFrame
import pandas as pd

from ..utility import all_dict_values_in_target_types
from ..watcher import get_obj_size
from .dtypes import dtype_auto_refine
from ..string.convert_functions import upper_and_replace_space_with_underscore

# Define a type for export mode which can be either 'stacked' or 'multi-column' while export DataFrame for PAR in DataSet
_export_mode = Literal["stacked","multi-column"]


# Define a custom dictionary class that behaves like a dictionary but allows attribute access
class liamDict(UserDict):
    # Allow attribute-style access to dictionary keys
    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError("liamDict object has no attribute '{key}'")
    
    # Allow attribute-style setting of dictionary keys
    def __setattr__(self, key, value):
        if key == 'data':   # <1>; Special handling for the 'data' attribute
            super().__setattr__(key, value)
            return
        self[key] = value
    
    # Override the setitem method to process keys and values before storing
    def __setitem__(self, key, value):
        key = self._get_key(key) # <2>
        value = self.handle_value(value) # <3>
        super().__setitem__(key, value) # <4>
    
    # Override the getitem method to process keys before retrieving
    def __getitem__(self, key):
        key = self._get_key(key)
        return super().__getitem__(key) # <5>
    
    # Process the key to avoid conflicts with class attributes or Python keywords
    def _get_key(self, key):
        if hasattr(self.__class__, key) or keyword.iskeyword(key): # <6>
            key += '_'
        return key

    # Recursively convert nested dictionaries to liamDict and process lists
    def handle_value(self, value):
        if isinstance(value, dict):
            value = liamDict(value)
        elif isinstance(value, list):
            for i, item in enumerate(value):
                value[i] = self.handle_value(item)
        return value

# Define a dictionary class that allows attribute access and provides additional methods
class attrDict(dict):
    # Allow attribute-style access to dictionary keys
    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError("attrDict object has no attribute '%s'" % key)

    # Allow attribute-style setting of dictionary keys        
    def __setattr__(self, key, value):
        self[key] = value
    
    # Return a JSON representation of the dictionary, excluding private keys
    def __str__(self):
        temp_dict = {"avaliable keys":[k for k in self.keys() if not k.startswith('_')]}
        # temp_dict.update({k:self[k] for k in self.keys() if k[0]=='_' and k[1]!='_' and (not k.startswith('_logger'))} )
        temp_dict.update({k:self[k] for k in self.keys() if k[0]=='_' and k[1]!='_' and not isinstance(self[k],logging.Logger) } ) 
        temp_json = json.dumps(temp_dict,indent=2)
        return temp_json

    # Use the string representation for the repr method
    def __repr__(self):
        return self.__str__()

    @property
    def _pureDict(self):
        return {k:self[k] for k in self.keys() if k[0]!='_' and not isinstance(self[k],logging.Logger)}
    
    # Export a subset of the dictionary based on a list of keys
    def subset(self,key_list:Iterable):
        """Create a subset of the attrDict with the specified keys

        Args:
            key_list (Iterable): list of some existing keys

        Returns:
            attrDict: result
        """
        result = {k: self[k] for k in self.keys() if k in key_list}
        return attrDict(result)

    # Refine the data types of all DataFrame objects in the dictionary to save memory
    def df_dtype_refine(self, inplace: bool = False, ignored_dfs:list=[])->object:
        """batch refine all Pandas dataframes by data types which save more memory

        Args:
            inplace (bool, optional): affects original data or not. Defaults to False.
            ignored_dfs (list, optional): list of keys that will be ignored by this method.

        Returns:
            object: copy of effective instance
        """
        if inplace:
            new_one = self
        else:
            new_one = copy.deepcopy(self)

        for k,v in self.items():
            if isinstance(v,DataFrame) and (k not in ignored_dfs):
                new_one[k] = dtype_auto_refine(v)
        return new_one

    # Standardize dictionary keys using a provided function
    def key_standardization(
        self, 
        standardize_func:callable = upper_and_replace_space_with_underscore
    ):
        """standardize the dictionary key by standardize_func

        Args:
            standardize_func (callable, optional): string conversion function. Defaults to toolbox.string.convert_functions.upper_and_replace_space_with_underscore.
        """
        original_keys = list(self.keys())
        for k in original_keys:
            self[standardize_func(k)]=self.pop(k)

    # Approximate the size of each item in the dictionary
    @property
    def size_summary(self)->dict:
        """iteratively approximate the size of each item in an attrDict instance

        Returns:
            Dict: result
        """
        rs = {k:get_obj_size(v) for k,v in self.items() if not k.startswith('_')}
        return rs

# Define a JSON encoder subclass to handle DataFrame objects
class JSE_4df(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, DataFrame):
            return obj.to_dict(orient='records')
        return json.JSONEncoder.default(self, obj)

# Define a class that represents properties of a dataset with type enforcement
class Property4DataSet(attrDict):
    # Initialize the class with type checking and assignment logic
    def __init__(self, value_type:type, input_dict:dict = None):
        _target_types = (
            datetime.date,
            datetime.datetime,
            datetime.time,
            str, 
            int, 
            float, 
            bool
            )

        if input_dict is None:
            pass
        elif value_type is None:
            
            if all_dict_values_in_target_types(input_dict, _target_types, visible_key_only=True):
                _ = [self.add(v, i) for i,v in input_dict.items() if not i.startswith('_')]
        elif isinstance(input_dict, Dict):
            if all_dict_values_in_target_types(input_dict, [List,], visible_key_only=True):
                input_dict = {i:v[0] for i,v in input_dict.items() if not i.startswith('_')}
            if all_dict_values_in_target_types(input_dict, [value_type,], visible_key_only=True):
                _ = [self.add(v, i) for i,v in input_dict.items() if not i.startswith('_')]
            else:
                raise TypeError(f'Not all values in attrDict are {value_type}')

    # Add a new property to the dataset
    def add(self,var, property_name):
        setattr(self,property_name, var)

    # Delete a property from the dataset    
    def delete(self,property_name):
        del self[property_name]

    # Return a list of all items excluding private and special attributes    
    @property    
    def allitems(self):
        return [x for x in list(self.keys()) if (x !=__name__) and (not x.startswith('_')) ] 
        #return [x for x in list(self.__dict__.keys()) if (x !=__name__) and (not x.startswith('_')) ] 

# Define a class to represent a dataset with predefined structures for dataframes, parameters, and mappings
class DataSet:
    __slots__ = (
        "DF",
        "PAR",
        "MAP"
    )
    # Initialize the dataset with optional dictionaries for each section
    def __init__(self, dict_PAR=None, dict_DF=None, dict_MAP=None):
        self.DF = self._DF(value_type = DataFrame, input_dict = dict_DF)
        self.PAR = self._PAR(value_type = None, input_dict = dict_PAR)
        self.MAP = Property4DataSet(value_type = dict, input_dict = dict_MAP)

    # Set or update the dataset sections with new dictionaries
    def set(self, dict_PAR:DataFrame=None, dict_DF:DataFrame=None, dict_MAP:DataFrame=None)->None:
        """assign dataframes to DataSet

        Args:
            dict_PAR (DataFrame, optional): DataFrame for PAR. Defaults to None.
            dict_DF (DataFrame, optional): DataFrame for Df. Defaults to None.
            dict_MAP (DataFrame, optional): DataFrame for MAP. Defaults to None.

        Raises:
            ValueError: when all Args are None
        """
        if all({(dict_PAR is None),(dict_DF is None),(dict_MAP is None)}): 
            raise ValueError("No attrDict instance has been assigned")
        if dict_DF:
            self.DF = self._DF(value_type = DataFrame, input_dict = dict_DF)
        if dict_PAR:
            self.PAR = self._PAR(value_type = None, input_dict = dict_PAR)
        if dict_MAP:
            self.MAP = Property4DataSet(value_type = dict, input_dict = dict_MAP)

    # Nested class for handling DataFrame properties
    class _DF(Property4DataSet):
        # Dump the DataFrame properties to a JSON file
        def dump2json(self, output_path: str = None):
            if not output_path:
                output_path = 'feed.DF.json'

            with open(output_path, 'w') as fp:
                json.dump(self, fp, cls=JSE_4df, indent = 2)

        # Standardize column names in all DataFrames
        def column_standardization(
                self,
                standardize_func:callable = upper_and_replace_space_with_underscore,
                inplace: bool = False
            ):
            """standardize the column name for each dataframe under DF by standardize_func

            Args:
                standardize_func (callable, optional): string conversion function. Defaults to upper_and_replace_space_with_underscore.
                inplace (bool, optional): Whether to modify the DataFrame rather than creating a new one. Defaults to False.

            Returns:
                Dataframe: it would be different than original DF property if `inplace=False`
            """
            if inplace:
                new_one = self
            else:
                new_one = copy.deepcopy(self)
           
            for df_name in new_one.allitems:
                col_names = list(new_one[df_name].columns)
                new_one[df_name].rename(
                    {col:standardize_func(col) for col in col_names},
                    axis='columns',
                    inplace=True
                )

            return new_one

    # Nested class for handling Parameter properties
    class _PAR(Property4DataSet):   
        # Export the parameters as a DataFrame
        def export_df(self, format: _export_mode = "multi-column"):
            """return a DataFrame by gathering all parameter:value within PAR  

            Args:
                format (_export_mode, optional): Literal["stacked","multi-column"]. Defaults to "multi-column".

            Returns:
                DataFrame
            """
            if format=="multi-column":
                df = pd.DataFrame(
                    [
                        [getattr(self,x) for x in self.allitems]
                    ],
                    columns = self.allitems
                    )
            if format=="stacked":
                df = pd.DataFrame(
                                zip(
                                    self.allitems,
                                    [getattr(self,x) for x in self.allitems]
                                ),
                                columns = ["parameter","value"]
                            )
            return df
    
# Define a class to represent a feed of data, which can be a simple or hierarchical structure
class Feed(DataSet):
    def __init__(self, multi_group = False, **kwargs):
        """Feed has a simple but organized structure to store the data

        Args:
            multi_group (bool, optional): Allows a hierarchical stucture. Defaults to False.
        """
        self.__multi_group = multi_group
        if multi_group:
            # If multi_group is True, provide a method to add new groups
            self.add_new_group = self.__add_new_group
        else:
            # Otherwise, initialize as a regular DataSet
            super().__init__(**kwargs)

    def __add_new_group(self, group_name, **kwargs):
        """Add a new group to the Feed if it's in multi-group mode.

        Args:
            group_name: The name of the new group.
            **kwargs: Additional arguments to pass to the DataSet initializer.
        """
        setattr(self, group_name, DataSet(**kwargs))

    # Get the names of all groups in the Feed if it's in multi-group mode.
    def __get_group_names(self):
        if self.__multi_group:
            return [g for g in list(self.__dict__.keys()) if g not in ('add_new_group','__add_new_group','__get_group_names','_Feed__multi_group')]
        else:
            return None

    def __str__(self):
        """String representation of the Feed, showing the data structure.

        Returns:
            A JSON-formatted string representation of the Feed's data.
        """
        if self.__multi_group:
            temp_dict = {
                g:[{x:[s for s in getattr(getattr(self,g),x).allitems]} for x in getattr(self,g).__slots__] 
                for g in self.__get_group_names()
            }
        else:
            temp_dict = {
                k:[x for x in getattr(self,k).allitems] 
                for k in self.__slots__
            }

        temp_json = json.dumps(temp_dict,indent=2)
        return "data attributes: \n" + temp_json
    
    # Representational string of the Feed, using the string representation.
    def __repr__(self):
        return self.__str__()





from functools import partial
from .convertor_methods_mapper import MAPPER


class Convertor:

    def __init__(self, obj):
        obj_type = type(obj)
        self.obj_type = f"{obj_type.__module__}.{obj_type.__name__}"
        self.input_obj = obj
        self._initialize_avai_methods()
        

    def _initialize_avai_methods(self):
        for k,v in MAPPER[self.obj_type].items():
            setattr(self,k,partial(v,self.input_obj))
            


import base64

import pandas as pd
import streamlit as st
from st_aggrid import AgGrid, GridOptionsBuilder,JsCode,GridUpdateMode, DataReturnMode
#from st_aggrid.shared import GridUpdateMode
from ...dao.connector import BlobConnector,parse_db_access
import pandas as pd
import azure



#db_access = parse_db_access("./examples/db_connector_testing/db.ini","optisuitestorage")

def ReadPictureFile(wch_fl):
    try:
        return base64.b64encode(open(wch_fl, 'rb').read()).decode()

    except:
        return ""

def create_BCE_explorer(df: pd.DataFrame,selected_rows:list):
    """Creates an st-aggrid interactive table based on a dataframe.

    Args:
        df (pd.DataFrame]): Source dataframe

    Returns:
        dict: The selected row
    """


    image_nation = JsCode("""function (params) {
        console.log(params);
        var element = document.createElement("span");
        var imageElement = document.createElement("img");
    
        imageElement.src = params.data.imgPath;
        imageElement.width="20";
        imageElement.height="20";

        element.appendChild(imageElement);
        element.appendChild(document.createTextNode(params.value));
        return element;
        }""")

    if df.shape[0] > 0:
        for _, row in df.iterrows():
            imgExt = row.imgPath[-4:]
            row.imgPath = f'data:image/{imgExt};base64,' + ReadPictureFile(row.imgPath)

    #theme: # options-> "streamlit","alpine", "light", "dark", "blue", "fresh", "material"
    #https://streamlit-aggrid.readthedocs.io/en/docs/GridOptionsBuilder.html
    gb = GridOptionsBuilder.from_dataframe(df)
    # gb = GridOptionsBuilder.from_dataframe(
    #     df, enableRowGroup=True, enableValue=True, enablePivot=True
    # )
    gb.configure_column('icon', cellRenderer=image_nation,width=80)
    gb.configure_column("imgPath", hide = "True")
    gb.configure_column("type", hide = "True")
    gb.configure_column("path", hide = "True")
    gb.configure_column("blob name",width=640)
    gb.configure_default_column(
        resizable=True,
        sortable=True,
        filterable =True,
        autoSizeColumns=True
    )
    #gb.configure_side_bar()

    grid_height = 300 #st.sidebar.number_input("Grid height", min_value=200, max_value=800, value=300)
    return_mode = "FILTERED" # st.sidebar.selectbox("Return Mode", list(DataReturnMode.__members__), index=1)
    return_mode_value = DataReturnMode.__members__[return_mode]

    selection_mode='single' #st.sidebar.radio("Selection Mode", ['single','multiple'], index=1)
    gb.configure_selection(selection_mode,
                            use_checkbox=True,
                            suppressRowClickSelection =True,
                            pre_selected_rows=selected_rows)
    #gb.configure_selection(selection_mode,use_checkbox=True,groupSelectsChildren=True, groupSelectsFiltered=True)
    #gb.configure_selection(selection_mode, use_checkbox=True, groupSelectsChildren=groupSelectsChildren, groupSelectsFiltered=groupSelectsFiltered)
    
    #gb.configure_pagination(paginationAutoPageSize=True)
    paginationPageSize=10
    gb.configure_pagination(paginationAutoPageSize=False,
                            paginationPageSize=paginationPageSize
                            )


    grid_response  = AgGrid(
        df,
        enable_enterprise_modules=False,
        gridOptions=gb.build(),
        width='60%',
        height=grid_height, 
        theme="streamlit",
        data_return_mode=return_mode_value, 
        update_mode=GridUpdateMode.SELECTION_CHANGED,
        allow_unsafe_jscode=True,
        fit_columns_on_grid_load=False,
        reload_data=True,
        key="selection",
        #columns_auto_size_mode=ColumnsAutoSizeMode.FIT_CONTENTS,
        #GridUpdateMode.MODEL_CHANGED,
    )

    return grid_response 

class BlobVirtualFolder:
    def __init__(self,db_access,path:str):
        BC = BlobConnector(db_access)
        self.path=path
        self.elements = pd.DataFrame(
            map(
                lambda x:['','folder',"folder.png",x.name,path] if isinstance(x,azure.storage.blob._list_blobs_helper.BlobPrefix) else ['','file',"file.png",x.name,path],BC.blob_container_client.walk_blobs(path)    
            ), columns = ['icon',"type","imgPath","blob name","path"]
        )

        self._folders = [
            x.name for x in BC.blob_container_client.walk_blobs(path)
            if isinstance(x,azure.storage.blob._list_blobs_helper.BlobPrefix)
        ]
        
        if path is not None:
            self.parent_path = self._get_parent_folder()
            self.elements = pd.concat([pd.DataFrame({
                "icon":'',
                "type":'folder',
                'imgPath':'undo.png',
                "blob name":self.parent_path,
                "path":path 
            },index =[0]),self.elements]).reset_index(drop = True)
        else:
            self.parent_path = "root"

    def _get_parent_folder(self):
        if len(self.path.split("/"))<=2:
            return None
        else:
            temp = self.path.split("/")
            return "/".join(temp[:-2])+"/"
    
    @property
    def folders(self):
        return self._folders

    @folders.setter
    def folders(self,f_list:list):
        self._folders = f_list

class BlobContainerExplorer:
    def __init__(self,db_access):
        self.BC = BlobConnector(db_access)
        self.db_access = db_access

        # initialize BVF and walk all folder by DFS
        self.BVF = self.start_dfs()
         
    @property
    def path(self):
        return self._path

    @path.setter
    def path(self,folder_path:str):
        self._path = folder_path
        self._parent_path=self.BVF[self._path].parent_path
        self._elements = self.BVF[self._path].elements
        

    @property
    def parent_path(self):
        return self._parent_path

    @parent_path.setter
    def parent_path(self):
        self._parent_path = self.BVF[self._path].parent_path

    @property
    def elements(self):
        return self._elements

    @elements.setter
    def elements(self):
        self._elements = self.BVF[self._path].elements


    def start_dfs(self):
        self.BVF = {}
        self.walk_folder(None)
        return self.BVF


    def walk_folder(self,path):
        if path not in self.BVF.keys():
            self.BVF[path] = BlobVirtualFolder(self.db_access,path)

        if len(self.BVF[path].folders)>0:
            self.walk_folder(self.BVF[path].folders.pop())
        elif self.BVF[path].parent_path != "root":
                return self.walk_folder(self.BVF[path].parent_path)
        else:
                return None





# import functools
# func = functools.partialmethod(pd.to_numeric,downcast="unsigned")
from pandas.core.frame import DataFrame
import pandas as pd


def dtype_auto_refine(df: DataFrame)->DataFrame:
    """This function will automatically convert numeric data to the most general type
    Also, replace object by pyarrow string for string columns.

    Args:
        df (DataFrame): Pandas Dataframe instance

    Returns:
        DataFrame: Dataframe with new data types assigned
    """
    df = df.copy(deep=True)
    individual_dtype_cols = {d_type: df.select_dtypes(include=[d_type]).columns for d_type in ["int","float","object"]}
    for d_type, cols in individual_dtype_cols.items():
        match d_type:
            case "int":
                df.loc[:,cols]=df.loc[:,cols].apply(pd.to_numeric,downcast="unsigned")
            case "float":
                df.loc[:,cols]=df.loc[:,cols].apply(pd.to_numeric,downcast="float")
            case "object":
                df.loc[:,cols] = df.loc[:,cols].astype("string[pyarrow]")           
    return df






import logging

from .db_connector import DBConnector

class NoSqlDBConnector(DBConnector):
    default_port = {
            "REDIS":6380,
            "MONGODB":7374,
        }

    def __init__(self, db_access, via_ssl = False, **kwargs):
        """connecting object for a NoSQL database

        Args:
            db_access (dict): return of parse_db_access()
            via_ssl (bool): connect via ssl
        """
        self.logger = logging.getLogger(__name__)
        self.cache_mode = 0
        self.cache_dir = "."

        
        self._db_access = db_access
        self._via_ssl = via_ssl

        self.queries_dir = None             
        self.__not_yet_purge = True

        self._db_type, self._port = self._check_nondefault_port()


# from .dtype_detect import dtype_detect
# from .ftype_detect import ftype_detect
from .loader import FileLoader

from .utility import (
    all_equal,
    count_folder_number,
    count_file_number,
    create_nonduplicated_dict_key,
)



import os
import __main__


class ToyboxApp:

    def __init__(self, script_path:str):
        """ to set a client streamlit script correctly  

        Args:
            script_path (str): path specified script
        """

        self.script_path = script_path
        self.script_dir = os.path.dirname(script_path) #it returns abspath
        if os.path.islink(self.script_dir):
            self.script_dir = os.readlink(self.script_dir) # if script_dir is a link, set script_dir to which the link points
        self.system_dir = os.path.dirname(__main__.__file__) 
        
        
    def set_working_dir(self):
        """ set working directory to correctly import local modules or load files

        Returns:
            dict: information
        """
        os.chdir(self.script_dir)
        return {
            "script_path":self.script_path,
            "script_dir":self.script_dir,
            "system_dir": self.system_dir
        }

    def get_realpath(self, filename:str)->str:
        """in case local files can't be read correctly when adding a project using softlink 

        Args:
            filename (str): file name

        Returns:
            str: real path of that file
        """
        return os.path.join(self.script_dir,filename)




import json
import pathlib
import os

from sqlalchemy import create_engine
from pandas import ExcelWriter
from ..files.utility import ensure_directory_exists
#from ...container import namedtuple_from_dict, BoundFuncAsClsMethod
from .utility import DataFrameJSONEncoder

def fl_to_json(self, output_path: str = './output.json',**kwargs):
    try:
        ensure_directory_exists(output_path)
        export_file_name = f'{output_path}'

        with open(export_file_name, 'w') as fp:
            json.dump(self._pureDict, fp, cls=DataFrameJSONEncoder, indent = 4,**kwargs)
        return True
    except Exception as e:
        raise e
        

def fl_to_csv(self, output_path: str = './output.csv', index=False,**kwargs):
    """
    Save a dictionary of DataFrames to separate CSV files, with each key appended to the filename.

    Args:
        self (FileLoader): instance of FileLoader
        output_path (str): The file path for the CSV files to save.
        index (bool): insert index to csv, default is False
    """
    
    try:
        ensure_directory_exists(output_path)
        base_file_name = f"{pathlib.Path(output_path).parent.__str__()}{os.sep}{pathlib.Path(output_path).stem}"
        for key, df in self._pureDict.items():
            csv_filename = f"{base_file_name}_{key}.csv"
            df.to_csv(csv_filename, index=index,**kwargs)
        return True
    except Exception as e:
        raise e

def fl_to_sqlite(self, output_path: str = './output.db', if_exists='replace',**kwargs):
    """
    Save a dictionary of DataFrames to an SQLite database file, with each key as a table name.

    Args:
        self (FileLoader): instance of FileLoader
        output_path (str): The file path of the SQLite database file to save.
        if_exists (str) : what to do when table exists. default is 'replace'
    """
    try:
        ensure_directory_exists(output_path)
        # Create a SQLite engine that represents the database connection
        engine = create_engine(f'sqlite:///{output_path}')
        
        with engine.connect() as connection:
            for table_name, df in self._pureDict.items():
                df.to_sql(table_name, con=connection, if_exists=if_exists, index=False,**kwargs)
        return True
    except Exception as e:
        raise e


def fl_to_excel(self, output_path: str = './output.xlsx',index=False, engine='openpyxl',**kwargs):
    """
    Save a dictionary of DataFrames to an Excel file, with each key as a worksheet name.

    Args:
        self (FileLoader): instance of FileLoader
        output_path (str): The file path of the Excel file to save.
        index (bool): insert index to csv, default is False
    """
    try:
        ensure_directory_exists(output_path)
        with ExcelWriter(output_path, engine=engine) as writer:
            for sheet_name, df in self._pureDict.items():
                df.to_excel(writer, index = index, sheet_name=sheet_name,**kwargs)
        return True
    except Exception as e:
        raise e



from json import JSONEncoder
from pandas import DataFrame

class DataFrameJSONEncoder(JSONEncoder):
    """
    Custom JSON Encoder subclass that converts pandas DataFrame objects to JSON.

    This encoder extends the default JSONEncoder and provides a custom
    serialization strategy for pandas DataFrame objects. When a DataFrame
    is encountered, it is converted to a dictionary with an orientation
    that lists each row as a dictionary (record).

    Any other object types will be handled by the default JSONEncoder
    implementation.
    """
    
    def default(self, obj):
        """
        Override the default method to serialize additional types.

        Args:
            obj (any): The object to serialize.

        Returns:
            any: A serializable object for obj, or calls the superclass
                 method for unsupported types.
        """
        if isinstance(obj, DataFrame):
            # Convert DataFrame to a list of dictionaries (one per row)
            return obj.to_dict(orient='records')
        # Use the default serialization for other types
        return JSONEncoder.default(self, obj)



import pathlib
import configparser
import os
from pdb import set_trace as bp
import logging
from typing import Dict, Literal, List
from functools import lru_cache
from collections.abc import Generator

import cx_Oracle
import pandas as pd
import numpy as np
from sqlalchemy import create_engine
from sqlalchemy.engine import URL
from sqlalchemy import text as sql_text

from ..dtype_mapping import map_db_datatypes_to_dtype, get_column_names_and_date_types

_cache_mode_options = Literal[0,1,2]


def parse_db_access(config_path: str, section_name:str)->Dict:
    """parse the configuration file to acquire required information for connecting supporting DBs

    Args:
        config_path (str): path of configuration file
        section_name (str): section name for db access in config file

    Returns:
        Dict: access information
    """
    logger = logging.getLogger(__name__)
    try:
        config = configparser.ConfigParser()
        _ = config.read(config_path)

        db_access = config._sections[section_name]
        db_access["ssl"] = config._sections["SSL"]
        # db_access = {}
        # db_access["server_username"] = config.get(section_name, 'username')

        if "db_type" not in db_access.keys(): raise ValueError("db_type is mandatory for setting db.ini")
        if db_access["db_type"]=="mssql" and "driver" not in db_access.keys():
            db_access["driver"] = check_odbc_driver()[0]

        return db_access

    except Exception as e:
        print(f"function parse_db_access got exception message: {e}")
        logger.info(
            "unable to parse from configuration file:/n {config_path} for database access information"
        )

def estimate_chunk_size(
        conn_str: str, 
        query: str,  
        cached_file_type:str='parquet.gzip', 
        compression_method:str='gzip',
        threshold_size:float=10.0, 
        sample_nrow:int = 2000,
        **kwargs
        )->int:
    """This function will estimate the idea chunk size which will result the output file around desired size

    Args:
        conn_str (str): connection string for creating database connection object; it has to meet the requirement for sqlalchemy 
        query (str): query statement
        cached_file_type (str, optional): intended file type in these options ['parquet.gzip','pickle','csv']. Defaults to 'parquet.gzip'.
        compression_method (str, optional): compression method for output file. Defaults to 'gzip'.
        threshold_size (float, optional): desired size for output file in MB. Defaults to 10.0.
        sample_nrow (int, optional): sample size in number of rows. Defaults to 2000.

    Returns:
        int: number of rows as chunk size

    For more information regarding conn_str, see https://docs.sqlalchemy.org/en/20/core/engines.html    
    """
    logger = logging.getLogger(__name__)

    executor = {
        'parquet.gzip':pd.DataFrame.to_parquet,
        'pickle':pd.DataFrame.to_pickle,
        'csv':pd.DataFrame.to_csv
    }

    if cached_file_type not in executor.keys():
        logger.error(f"Receive unsupported value {cached_file_type} for argument cached_file_type; current supportting list:\n {executor.keys()}")
        raise ValueError(f"Receive unsupported value {cached_file_type} for argument cached_file_type; current supportting list:\n {executor.keys()}")
    
    eng = create_engine(conn_str)
    conn = eng.connect() 
    query = sql_text(query)
    logger.info(f"estimate chunk size for query {query} targeting on a threshold file size= {threshold_size}MB for {cached_file_type} using {sample_nrow} rows as sampling size")

    try:
        result = conn.execute(query)
        chunk = result.fetchmany(sample_nrow)
        actual_number_of_rows = len(chunk)
        sample_df = load_query_result_into_df_with_intended_dtypes(
                chunk,
                result
            )

    except Exception as e:
        logger.debug(f"not able to get sample dataframe for estimating the chunk size via following query \n {query} \n , got error message: {e}")
        conn.close()
    finally:
        conn.close()

    # Convert the sample data to desired format and get its size
    sample_file_path = f'sample.{cached_file_type}'
    executor[cached_file_type](sample_df, f'sample.{cached_file_type}', compression=compression_method)
    sample_file_size = os.path.getsize(sample_file_path)

    # Define the intended threshold file size in bytes (e.g., 15 MB)
    intended_threshold_size = threshold_size * 1024 * 1024  # threshold_size in bytes

    # Calculate the estimated chunk size based on the sample file size; be noticed that actual number of rows may be less than sample size
    if actual_number_of_rows< sample_nrow:
            sample_nrow = actual_number_of_rows
    estimated_chunk_size = int((intended_threshold_size / sample_file_size) * sample_nrow)

    # Cleanup: Remove the sample file
    os.remove(sample_file_path)
    logger.info(f"estimated chunk size is {estimated_chunk_size}")

    return estimated_chunk_size


def read_sql_query_in_chunks_sqlalchemy(query, engine, chunk_size:int, enforce_dtype:bool)->Generator:
    """
    Retrieve data from a database in assigned-sized chunks.

    Parameters:
        query: SQL query to execute.
        engine: SQLAlchemy engine object.
        threshold: Maximum number of rows to fetch in each chunk.

    Returns:
        A generator that yields DataFrames for each chunk of data.
    """
        
    # Create a connection
    conn = engine.connect() 
    
    # Execute the query
    query = sql_text(query)
    result = conn.execute(query)

    # Get the column names from the cursor description
    column_names = result.keys()

    # Initialize variables for chunk iteration
    chunk = result.fetchmany(chunk_size)

    while chunk:
        if enforce_dtype:
            chunk_data = load_query_result_into_df_with_intended_dtypes(
                chunk,
                result
            )
        else:# Create a DataFrame from the fetched chunk
            chunk_data = pd.DataFrame(chunk, columns=column_names)
        
        # Convert numeric columns to either int or float
        #chunk_data = only_int_float_for_numeric_columns(chunk_data)

        yield chunk_data

        # Fetch the next chunk
        chunk = result.fetchmany(chunk_size)


def load_query_result_into_df_with_intended_dtypes(rs, result, mapper:dict = None)->pd.DataFrame:
    """Turn query record set into Pandas dataframe with intended dtype

    Args:
        rs : record set, i.e.,sequence[row]. Returns of sqlalchemy result.fetchall() or .fetchmany() 
        result : sqlalchemy CursorResult
        mapper (dict): map out database data type and dtypes for dataframe

    Returns:
        pd.DataFrame
    """
    if mapper is None:
        mapper = map_db_datatypes_to_dtype
    column_names, db_data_types = get_column_names_and_date_types(result)

    database_type = result.connection.dialect.name
    database_driver = result.connection.engine.driver

    # create intended dtype dict via iterating over column names and mapping the database data type
    intended_dtypes = {col_name:mapper(database_type,database_driver,db_data_type) for col_name, db_data_type in zip(column_names, db_data_types)}

    return pd.DataFrame(rs,columns=column_names).astype(intended_dtypes)

def read_sql_query_in_chunks_oracle(query:str, connection, chunk_size=20000):
    """
    Retrieve data from an Oracle database in adaptive-sized chunks using cx_Oracle.

    Parameters:
    - query: SQL query to execute.
    - connection: cx_Oracle Database connection object; Establish a connection to the Oracle database by cx_Oracle.connect(connection_string) and the format of connection_string is "user/password@hostname:1521/your_service_name"
    - threshold: Maximum number of rows to fetch in each chunk (default is 20,000).

    Returns:
    - A generator that yields DataFrames for each chunk of data.
    """

    # Create a cursor
    cursor = connection.cursor()

    # Execute the query
    cursor.execute(query)

    # Get the column names from the cursor description
    column_names = [desc[0] for desc in cursor.description]

    # Initialize variables for chunk iteration
    chunk = cursor.fetchmany(chunk_size)

    while chunk:
        # Create a DataFrame from the fetched chunk
        chunk_data = pd.DataFrame(chunk, columns=column_names)

        yield chunk_data

        # Fetch the next chunk
        chunk = cursor.fetchmany(chunk_size)



def check_odbc_driver():
    from pyodbc import drivers
    # Get a list of installed ODBC drivers
    odbc_drivers = drivers()
    found_drivers=[]
    # Check for SQL Server drivers and print their names
    sql_server_drivers = [driver for driver in odbc_drivers if 'SQL Server' in driver]
    if sql_server_drivers:
        
        for driver in sql_server_drivers:
            if "ODBC" in driver:
                found_drivers.append(driver)
    else:
        raise Exception("No SQL Server ODBC drivers are found. Please set availavble driver in db.ini file for your SQL Server")

    return found_drivers


class DBConnector:
    default_port = {
            "ORACLE":1521,
            "MSSQL":1433,
            "AZURE-BLOB":None,
            "MARIADB":3307,
            "SNOWFLAKE":443
        }

    def __init__(self, db_access, via_ssl = False, **kwargs):
        """connecting object for a database

        Args:
            db_access (dict): return of parse_db_access()
            via_ssl (bool): connect via ssl
        """
        self.logger = logging.getLogger(__name__)
        self.cache_mode = 0
        self.cache_dir = "."

        
        self._db_access = db_access
        self._via_ssl = via_ssl
        if self._via_ssl:
            if "ssl" not in self._db_access.keys():
                self.logger.error("Please check db.ini and ensure secton SSL is set correctly while connecting via_ssl")

        self.queries_dir = None             
        self.__not_yet_purge = True

        self._db_type, self._port = self._check_nondefault_port()
            
    def set_cache_dir(self, cache_dir: str, cache_mode: _cache_mode_options = 1)->None:
        """set up path for directory where cached files are stored

        Args:
            cache_mode (int, optional): 0: always pull data from db
                                1: only pull data if cache doesn't exist
                                2: refresh cache before using 
                                Defaults to 0.
            cache_dir (str): directory for keeping/searching cache files; only works if cache_mode in {1,2}
        """

        self.cache_dir = cache_dir
        self.cache_mode = cache_mode
        

    def _check_nondefault_port(self)-> List:
        db_type = self._db_access["db_type"].upper()
        
        if "port" in self._db_access.keys():
            port = self._db_access["port"]
        else:
            port = self.default_port[db_type]

        return [db_type, port]
        
    def set_queries_dir(self,queires_dir: str, **kwargs)->None:
        """set up path for directory where predefined sql statement files are stored

        Args:
            queires_dir (str): path of directory
        **kwargs:
            all variables within predefiend sql queries can pass to DBConnector via keyword arguments
        """
        self.queries_dir = queires_dir
        self.query_args = dict(kwargs)

    def del_cache(self)->None:        
        """delete all cache files in cache directory
        """     
        dir = self.cache_dir
        for f in os.listdir(dir):
            if f.endswith(".pkl"):
                os.remove(os.path.join(dir, f))

    def _check_cache_dir(self)->None:
        if not os.path.exists(self.cache_dir):
            os.makedirs(self.cache_dir)

    def _correct_datetime_format_for_mssql(self):
        for k,v in self.query_args.items():
            if type(v)==np.datetime64: 
                self.query_args[k] = str(v.astype('M8[D]'))

    
    def create_engine(self,**kwargs):
        if self._via_ssl:
            return create_engine(self._conn_str, connect_args = self._ssl_args,**kwargs)
        else:
            return create_engine(self._conn_str,**kwargs)
    
    def _get_mssql_conn_str(self,):
        driver_name = self._db_access["driver"]
        match driver_name:
            case str() as s if "ODBC" in s:
                conn_str=''.join(
                    [
                    'DRIVER={'+ driver_name +'};',
                    f'SERVER={self._db_access["server"]},{self._port};',
                    f'DATABASE={self._db_access["database_name"]};UID={self._db_access["server_username"]};PWD={self._db_access["server_password"]}'
                    ]
                )
                conn_str = URL.create("mssql+pyodbc", query={"odbc_connect": conn_str})

            case "FreeTDS":
                conn_str = URL.create(
                    'mssql+pyodbc',
                    username=self._db_access["server_username"],
                    password=self._db_access["server_password"],
                    host=self._db_access["server"],
                    port=self._port,
                    database=self._db_access["database_name"],
                    query={
                        "driver": "FreeTDS",
                        "TrustServerCertificate": "yes"  
                    })
        return conn_str

    def _get_conn_str_for_cx_Oracle(self):
        return f"{self._db_access.server_username}/{self._db_access.server_password}@{self._db_access.server}:{self._port}/{self._db_access.service_name}"
        
    def _get_oracle_conn_str(self):
        return f'oracle+cx_oracle://{self._db_access["server_username"]}:{self._db_access["server_password"]}@{self._db_access["server"]}:{self._port}/?service_name={self._db_access["service_name"]}'

    def _get_snowflake_conn_str(self):
        return URL.create(
            'snowflake',
            username = self._db_access["user"],
            password = self._db_access["password"],
            host = self._db_access["account"] + '.snowflakecomputing.com',
            #database = 'SUPPLY_CHAIN',
            query = {
                #"role":'if so',
                "warehouse": self._db_access["warehouse"],
                "account": self._db_access["account"].split("-")[1].upper(),
                #"authenticator":'externalbrowser'
            }
        )

    
    def _get_mariadb_conn_str(self):
        return f"mysql+pymysql://{self._db_access['user']}:{self._db_access['password']}@{self._db_access['host']}:{self._db_access['port']}/{self._db_access['database']}"

    @property
    def _ssl_args(self):    
        return {"ssl":self._db_access["ssl"]}

    @property
    def _conn_str(self):

        conn_str = {
            "MSSQL": self._get_mssql_conn_str,
            "ORACLE": self._get_oracle_conn_str,
            "MARIADB": self._get_mariadb_conn_str,
            "SNOWFLAKE": self._get_snowflake_conn_str
        }

        return conn_str[self._db_type]()

    def load_args_for_predefined_query(
            self, 
            query_file: str,
            query_args:dict=None
            )->str:
        """This method returns the query statement for pulling data by filling in query arguments for predefined query

        Args:
            query_file (str): name of predifined query file (extension should be .sql)
            query_args (dict, optional): dictionary of query arguments. Defaults to None.

        Returns:
            str: sql query statement
        """
        
        if query_args is None:
            query_args = self.query_args

        if  self.queries_dir not in query_file:
            query_file = os.path.join(self.queries_dir, f"{query_file}.sql")

        with open(query_file, 'r') as fname:
            query = fname.read()
            query_str = query.format(
            **query_args
            )
        return query_str

    def dump_to_db(
            self,
            df,
            dest_tb_name,
            is_fast_executemany = True,
            index = False,
            index_label=None,
            if_exists = 'append',
            dtype = None,
            chunksize = None,
            method=None,
    ):
        """dump a dataframe to a specific table

        Args:
            df (_type_): _description_
            dest_tb_name (_type_): _description_
            is_fast_executemany (bool, optional): _description_. Defaults to True.
            is_index (bool, optional): add index automatically. Defaults to False.
            if_exists (str, optional): what to do while data exists in table. 'fail', 'replace', 'append'. Defaults to 'append'.
                                    How to behave if the table already exists.
                                    fail: Raise a ValueError.
                                    replace: Drop the table before inserting new values.
                                    append: Insert new values to the existing table.
            dtype (dict, optional): dict to dedicate the dtype for each columns. Defaults to None.
                                    example dtype:
                                    using df column names as keys
                                    {
                                    'datefld': sqlalchemy.DateTime(), 
                                    'intfld':  sqlalchemy.types.INTEGER(),
                                    'strfld': sqlalchemy.types.NVARCHAR(length=255)
                                    'floatfld': sqlalchemy.types.Float(precision=3, asdecimal=True)
                                    'booleanfld': sqlalchemy.types.Boolean
                                    }   
        """
        try: 
            # conn_str = self._conn_str
            # eng = create_engine(conn_str)

            eng = self.create_engine()

            with eng.connect() as conn:
                df.to_sql(dest_tb_name, conn, if_exists = if_exists, index = index, index_label=index_label, chunksize=chunksize, dtype=dtype,method=method)
        
        except Exception as e:
            self.logger.error(e) 


    def exec_SQL(self,statement):
        """execute general sql statement

        Args:
            statement (str): SQL statement

        Returns:
            CursorResult: instance of sqlalchemy.engine.cursor.CursorResult
        """
        eng = self.create_engine()
        try:
            with eng.connect() as conn:
                rs = conn.execute(sql_text(statement))
                conn.commit()
                return rs
        except Exception as e:
            self.logger.error(e)

    def pull_SQL(
            self,
            query,
            chunk_size:int=None,
            **kwargs
    ):
        """simply pulling data with SQL query; efficient pulling by chunks should consider using method pull_predefined_query  
        Args:
            query (str): sql statement
            chunk_size (int): pull by chunk if assigned (optional); using Pandas implementation for pulling by chunks
        Return:
            df (DataFrame): results of executing SQL statement if it has return and chunk_size<2 or None
            dfs_generator (Generator): generator for all resulted chunks if chunk_size>1
        """
        try:
            # conn_str = self._conn_str
            # eng = create_engine(conn_str)

            eng = self.create_engine()

            with eng.connect() as conn:
                if chunk_size in (None,1):
                    df = pd.read_sql(
                        sql_text(query),
                        conn,
                    )
                    return df
                else:
                    dfs_generator = pd.read_sql(
                        sql_text(query),
                        conn,
                        chunksize=chunk_size
                    )
                    return dfs_generator
        except Exception as e:
            self.logger.error(e)       

    @staticmethod
    def dump_dfs_generator(dfs_generator, keys:list=None):
        """Dump DataFrames from a generator into a dictionary.

        Args:
            dfs_generator (generator): A generator that yields DataFrames.
            keys (list or None, optional): List of keys for the generated dictionary.
                If None, numerical indices will be used as keys. Default is None.

        Returns:
            dict: A dictionary containing DataFrames from the generator, with keys as specified.

        Example:
        ```
        # Generate sample DataFrames
        num_dataframes = 5
        rows_per_dataframe = 10
        dataframes = generate_dfs(num_dataframes, rows_per_dataframe)

        # Test the dump_dfs_generator method
        keys = ['DF1', 'DF2', 'DF3', 'DF4', 'DF5']  # Example keys
        result = dump_dfs_generator(dataframes, keys)
        print(result)
        ```

        This function takes a generator yielding DataFrames and returns a dictionary
        with specified keys (or numerical indices if keys are not provided), where each
        key corresponds to a DataFrame yielded by the generator.
        """

        if keys is None:
            return {i: chunk_df for i, chunk_df in enumerate(dfs_generator)}
        else:
            return {i: chunk_df for i, chunk_df in zip(keys,dfs_generator)}

    def pull_predefined_query(
        self,
        query_name,
        chunk_mode:bool=False,
        **kwargs
    ):
        """pull data by predefined sql statement file including variables.

        Args:
            query_name (str): name of a predefined sql statement file; no extension needed
            chunk_mode (bool): pull data by chunk mode or not
            
            kwargs: 
                    - chunk_size (int): number of rows for a chunk; estimate_chunk_size() will be trigger when chunk_size is not assigned. Therefore, more details for accepted kwargs could check description of function estimate_chunk_size()
                    - capitalize_column_name (bool): convert all column name to uppercase or not
                    - enforce_dtype (bool): convert database data types to intended dtypes for resulted dataframe; only enables for chunk_mode
        Raises:
            Exception: queries_dir can't be empty, set by method set_queries_dir() first
            ValueError: if <query_name>.sql can't be found in queries_dir

        Returns:
            DataFrame: result of executing SQL statement
        """
        if self.queries_dir is None:
            self.logger.error("Use method set_queries_dir(queries_dir:str) to enable method pull_predefined_query") 
            raise Exception("Use method set_queries_dir(queries_dir:str) to enable method pull_predefined_query")
        
        if f"{query_name}.sql" not in os.listdir(self.queries_dir):
            self.logger.error('Query for {} is undefined'.format(query_name))
            raise ValueError('Query for {} is undefined'.format(query_name))
        
        executor = {
            "do_chunk":self._run_chunk_mode,
            0: self._run_cache_mode_0,
            1: self._run_cache_mode_1,
            2: self._run_cache_mode_2,
        }

        if chunk_mode:    
            dfs_generator = executor["do_chunk"](query_name,**kwargs)
            return dfs_generator
        else:
            df = executor[self.cache_mode](query_name,**kwargs)
            if "capitalize_column_name" in kwargs and kwargs["capitalize_column_name"]:
                df.columns = df.columns.str.upper()
            
            return df

    def _run_chunk_mode(self, query_name:str, chunk_size:int=None, enforce_dtype: bool=False, **kwargs)->Generator:

        conn_str = self._conn_str

        query_file = os.path.join(self.queries_dir, f"{query_name}.sql")
        query_str= self.load_args_for_predefined_query(query_file,{**self.query_args,**kwargs})
     
        if chunk_size is None:
            eng = self.create_engine()
            chunk_size = estimate_chunk_size(
                conn_str,
                query_str,
                **kwargs
            )

        eng = self.create_engine()
        
        dfs_generator = read_sql_query_in_chunks_sqlalchemy(
                            query = query_str,
                            engine = eng,
                            chunk_size = chunk_size,
                            enforce_dtype = enforce_dtype
                        )

        return dfs_generator

        
    def _run_cache_mode_0(self,query_name, **kwargs):
        query_file = os.path.join(self.queries_dir, f"{query_name}.sql")
        query_str= self.load_args_for_predefined_query(query_file,{**self.query_args,**kwargs})

        try:
            df = self.pull_SQL(query_str)
        except Exception as e:
            self.logger.info(f"predefined query {query_name} got exception message: {e}")
        return df

    def _run_cache_mode_1(self,query_name, **kwargs):
        pickle_path = os.path.join(self.cache_dir, f"{query_name}.pkl")
        if f"{query_name}.pkl" in os.listdir(self.cache_dir):  
            df = pd.read_pickle(pickle_path)
        else:
            df = self._run_cache_mode_0(query_name)
            df.to_pickle(pickle_path)
        return df

    def _run_cache_mode_2(self,query_name, **kwargs):
        if self.__not_yet_purge: 
            self._del_cache()
            self.__not_yet_purge = False

        df = self._run_cache_mode_1(query_name)
        return df

if __name__ == "__main__":
    pass 


import os
import datetime
import pickle
import time
import logging
from logging.config import fileConfig
import logging
from pdb import set_trace as bp
from decimal import Decimal
from typing import (
    Literal,
    Optional
)

import numpy as np
import pandas as pd
import IPython


LOGGING_LEVEL = Literal['DEBUG','INFO','WARNING','ERROR','CRITICAL']

def file_last_update_datetime(filepath:str, datetime_only:bool=False):
    modTimesinceEpoc = os.path.getmtime(filepath)
    modificationTime = datetime.datetime.utcfromtimestamp(modTimesinceEpoc).strftime('%Y-%m-%d %H:%M')
    if datetime_only:
        return modificationTime
    else:
        return 'Last Update: \n' + modificationTime + ' UTC'


def all_dict_values_in_target_types(
    input_dict: dict,
    target_types: list,
    visible_key_only = True
) -> bool:
    """check if all values of input dictionary are in target types

    Args:
        input_dict (dict): dictionary wait to check
        target_types (list): list of types
        visible_key_only (bool, optional): only check dictionary keys not starting from underscore. Defaults to True.

    Returns:
        bool : True or False
    """
    if visible_key_only:
        return all(type(v) in target_types for k,v in input_dict.items() if not k.startswith('_'))
    else:
        return all(type(v) in target_types for k,v in input_dict.items())
    


def from_cache_if_possible(df_name:str, cache_dir:str, refresh_func:callable,**kwarg):
    """provide basic functionality to load pickle file as a variable if it exists; otherwise, call refresh_func to return the data

    Args:
        df_name (str): variable name
        cache_dir (str): path for checking the existance of pickle file
        refresh_func (callable): function to return the intended data

    Returns:
        obj: instance of intended data
    """
    cache_path=os.path.join(
        cache_dir,
        f"{df_name}.pkl"
    )
    if os.path.exists(cache_path):
        file = open(cache_path, 'rb')
        df = pickle.load(file)
        file.close()
        #df = pd.read_pickle(cache_path)
    else:
        df = refresh_func(**kwarg)
        file = open(cache_path, 'wb')
        pickle.dump(df, file)
        file.close()
        #pd.to_pickle(df, cache_path)
    return df



def adjust_decimal_types(df, precision:int=8):
    precision_str = '0.' + '0' * (precision - 1) + '1'  # Construct precision string
    for col in df.select_dtypes(include=[np.number], exclude=[np.bool_]).columns:
        df[col] = df[col].apply(lambda x: round(x, precision) if isinstance(x, float) else Decimal(x).quantize(Decimal(precision_str)))

    return df



def set_logger(log_level: LOGGING_LEVEL = 'INFO', use_file_config: bool = False, config_file: Optional[str] = None):
    """quick solution for setting/initializing a logger
    For a basic setup without a configuration file such as for Jupyter notebook:
            set_logget(log_level='DEBUG')
    For setting up using a configuration file:
            set_logger(use_file_config=True, config_file='logging.ini')

    Args:
        log_level (LOGGING_LEVEL, optional): level of log. Defaults to 'INFO'.
        use_file_config (bool, optional): use configuration file or not. Defaults to False.
        config_file (Optional[str], optional): path of configuration file. Defaults to None.

    Returns:
        logger


    """
    logger = logging.getLogger()
    logger.setLevel(log_level)

    handler = logging.StreamHandler()
    handler.setLevel(log_level)

    formatter = logging.Formatter('%(levelname)s - %(message)s')
    handler.setFormatter(formatter)

    logger.addHandler(handler)

    if use_file_config and config_file:
        fileConfig(config_file, disable_existing_loggers=False)

    return logger


def autosave_notebook():
    IPython.display.Javascript('IPython.notebook.save_notebook()')


import sqlite3
import logging
import pathlib

import filetype
from filetype.types.archive import *

# determine file type via Magic number may sometime be tricky 
# so the following extensions are reported by developer community. 
# Depends on the way you generate these files, fieltype won't always determine correctly.
# For those internal programs, we should just trust the file ext 
# really tells the right file type, and the program will focus on determine the compression algorithm
 
warning_list = [
    'xlsx',
    'xlsb',
    'csv',
    'json'
]

def get_file_extension(file_path:str) -> str:
    return pathlib.Path(file_path).suffix

def ftype_detect(file_path) -> str:
    logger = logging.getLogger(__name__)
    file_ext = get_file_extension(file_path)
    if file_ext in warning_list:
        return file_ext
    else:
        ftype = filetype.guess(file_path)
        file_ext =filetype.guess_extension(file_path)

        if type(ftype) not in ['Tar','Gz','Zip']:
            return ftype,filetype.guess_extension(file_path)

    print('File extension: %s' % ftype.extension)
    return filetype.guess_extension(file_path)
    ftype = filetype.guess(file_path)
    if ftype is None:
        logger.info(f"can't determine the file type for {file_path}")
        return None
    else:
        logger.info(f"file type {ftype} is detected for {file_path}")
        return ftype
    
    try:
        con = sqlite3.connect(file_path)
        cur = con.cursor()
        cur.execute("PRAGMA integrity_check")
        ftype = 'sqlite'
        return ftype
    except sqlite3.DatabaseError:
        con.close()
    
    try:
        pass
    except ValueError:
        return ftype
        
        
    file_ext = pathlib.Path(file_path).suffix
    ftype = file_ext[1:]

    return ftype

if __name__ == '__main__':
    ftype_detect()




from pdb import set_trace as bp
import json
import shutil
import tempfile
import pathlib
import glob
import logging


import pandas as pd
import dask.dataframe as dd
from pandas.core.frame import DataFrame


from ...dao.feed import attrDict
from .utility import (
    all_equal,
    count_folder_number,
    count_file_number,
)


# Exception class
class StructError(ValueError):
    """Raised when the composition of an archive is not supported by package"""
    _default_msg = "invalid composition of the archive"
    def __init__(self, message = None):
        self.message = self._default_msg if message is None else message
    def __str__(self):
        return self.message

class FormatError(StructError):
    def __init__(self, message = None):
        self.message = f"only support file format in {FileLoader.supported_format}" if message is None else message


class InconsistentError(StructError):
    def __init__(self, message = None):
        self.message = f"for multiple files, only support .csv, .pkl or .parquet" if message is None else message

def get_file_extension(file_path:str) -> str:
    return pathlib.Path(file_path).suffix

def ftype_detect(file_path:str) -> str:
    # TODO: additional checking for sqlite, json, and pkl
    return get_file_extension(file_path)

    # file_ext = get_file_extension(file_path)
    # if file_ext in warning_list:
    #     return file_ext
    # else:
    #     ftype = filetype.guess(file_path)
    #     file_ext =filetype.guess_extension(file_path)

    #     if type(ftype) not in ['Tar','Gz','Zip']:
    #         return ftype,filetype.guess_extension(file_path)

    # print('File extension: %s' % ftype.extension)
    # return filetype.guess_extension(file_path)
    # ftype = filetype.guess(file_path)
    # if ftype is None:
    #     logger.info(f"can't determine the file type for {file_path}")
    #     return None
    # else:
    #     logger.info(f"file type {ftype} is detected for {file_path}")
    #     return ftype
    
    # try:
    #     con = sqlite3.connect(file_path)
    #     cur = con.cursor()
    #     cur.execute("PRAGMA integrity_check")
    #     ftype = 'sqlite'
    #     return ftype
    # except sqlite3.DatabaseError:
    #     con.close()
    
    # try:
    #     pass
    # except ValueError:
    #     return ftype
        
        
    # file_ext = pathlib.Path(file_path).suffix
    # ftype = file_ext[1:]

    # return ftype

def update_target_dir(target_dir:str) -> str:
        """check target directory and return the correct path if subfolder exists

        Args:
            target_dir (str): path of target directory

        Raises:
            StructError: invalid composition of the archive

        Returns:
            str: path of target directory
        """
        #logger.info(f"check if subfolder exists in archive")
        contents = get_contents_list(target_dir)
        num_folder = count_folder_number(contents)
        num_file = count_file_number(contents)

        if num_folder>1: #more than two subfolder within original temp dir
            #logger.info(f"[Error]: more than one subfolder is found the archive {target_dir}")
            raise StructError(f"more than one subfolder is found in target directory {target_dir}")
        elif num_folder==1: #just one subfolder in original temp_dir
            if num_file>0: 
                #logger.info(f"[Error]: either single subfolder or files is allowed in target directory {temp_dir}")
                raise StructError("subfolder can't stand with files")
            else:
                target_dir = contents[0]  
        elif num_file==0: #no subfolder within temp dir
            raise ValueError("no contents in target directory or directory doesn't exist")
        return target_dir

def get_contents_list(path:str, recursive_search = False) -> list[str]:
    """Return contents under a given directory

    Args:
        path (str): path of a directory
        recursive_search (bool, optional): search recursively if subfolders exist. Defaults to False.

    Returns:
        list[str]: list of contents
    """
    if recursive_search:
        contents = glob.glob(f"{path}/**", recursive=True)[1:] #exculde path dir itself
    else:
        contents = glob.glob(f"{path}/**", recursive=False) 
    return contents


def load_json2ad(file_path):
    dicts_from_json = json.load(open(file_path))
    dfs2dict = {
        key: pd.DataFrame(dicts_from_json[key]) for key in dicts_from_json
    }
    return dfs2dict



class FileLoader(attrDict):
    """_summary_

    Supported format:
        excel(xlsx): default is to load all worksheets
        csv: separator can be assign to method `execute` as arg `sep` and its default is `','; note that `sep=None` means automatically detect 
    """
    supported_format = (     
        '.xlsx',
        '.xlsb',
        '.csv',
        '.json',
        '.db',
        '.parquet',
        '.pkl'
    )

    load_method = {
            ".csv":pd.read_csv,
            ".xlsx":pd.read_excel,
            ".parquet": pd.read_parquet,
            ".pkl": pd.read_pickle,
            ".json": load_json2ad
        }

    __slots__ = (
        "_input_path",
        "_archive",
        "_logger",
    )

    def __init__(self, path: str, archive = False):
        """Load source data to a dictionary.  
           path can only point to a file or directory(includes multiple identical files).
           Supporting file types:  
                - comma-separated values (.csv)
                - Excel (.xlsx, .xlsb)
                - Sqlite (.db, .sqlite)
                - JSON (.json)
                - Apache Parquet (.parquet)
                - Pytho pickle file (.pkl)

        For loading multiple files, only supports csv, pkl, or parquet.
        path can also point to an archive of file(s) of a folder, the compress method can be:  
                - tar.gz
                - tar.bz2
                - tar.xz
                - zip

        Args:
            path (str): a path to a file or directory
            archive (bool, optional): if an archive file is the target of the path. Defaults to False.

        All class properties should start with _ instead of those dataframes
        """
        self._input_path = path
        self._archive = archive 
        self._logger = logging.getLogger(__name__)

    def _parse_path(self):
        ftype = None
        if pathlib.Path(self._input_path).is_dir():
            path_is_for_dir = True
        elif pathlib.Path(self._input_path).is_file():
            path_is_for_dir = False
            ftype = ftype_detect(self._input_path)
        else:
            raise Exception("path can only be an existing file or path")
            
        return path_is_for_dir, ftype

    @property
    def _pureDict(self):
        return {k:self[k] for k in self.keys() if k not in self.__slots__ and not isinstance(self[k],logging.Logger)}


    def execute(self, **kwargs):
        path_is_for_dir, ftype = self._parse_path()

        if path_is_for_dir:
            self._load_when_path_is_folder(archive=False,**kwargs)

        elif ftype is not None:
            if self._archive:
                self._load_when_path_is_folder(archive=True, **kwargs)
            elif ftype in self.supported_format:
                self._load_to_attrDict([self._input_path,],True,ftype, **kwargs)

        else:
            exception_msg = "path should only point to a directory or a file (could be an archive with optional argument archive = True"
            self._logger.info(exception_msg)
            raise TypeError(exception_msg)
    

    def _load_to_attrDict(self,contents:list[str], single_file:bool, file_type:str, **kwargs):

        if single_file:
            match file_type:
                case '.xlsx':
                    _ = [self.update({k:v for k,v in  FileLoader.load_method[file_type](contents[0], sheet_name=None, **kwargs).items()})]
                case other:
                    self.update({pathlib.Path(contents[0]).stem.replace(' ', '_'):FileLoader.load_method[file_type](contents[0],**kwargs)})
        else:
            _ = [self.update({pathlib.Path(c).stem.replace(' ', '_'): FileLoader.load_method[file_type](c,**kwargs)}) for c in contents]


    def _load_when_path_is_folder(self, archive, **kwargs):
        temp_dir = None
        if archive:
            temp_dir = tempfile.TemporaryDirectory()
            target_dir = ""

            try:
                shutil.unpack_archive(self._input_path,temp_dir.name)
            except shutil.ReadError:
                self._logger.info(f"can't unpack the achive, check input file {self._input_path}")
                raise shutil.ReadError(f"can't unpack the achive, check input file {self._input_path}")

            target_dir = update_target_dir(temp_dir.name)
        else:
            target_dir = update_target_dir(self._input_path)

        self._load_feed_from_source_dir(target_dir)

        if temp_dir is not None: temp_dir.cleanup()

        


    def _load_feed_from_source_dir(self,target_dir:str, **kwargs) -> attrDict:
        """load data from files in target directory to an instance of attrDict

        Args:
            target_dir (str): path of target directory

        Raises:
            StructError: invalid composition of the archive
            InconsistentError: for multiple files, only support .csv, .pkl or .parquet
            FormatError: unsupported file type
            ValueError: no contents in target directory or directory doesn't exist

        Returns:
            attrDict: collections of input data
        """

        contents = get_contents_list(target_dir)
        num_folder = count_folder_number(contents)
        num_file = count_file_number(contents)
        file_type= ftype_detect(contents[0])
        identical_ftype = all_equal(get_file_extension(c) for c in contents)

        if num_folder>0:
            raise StructError("nested subfolder is not allowed")
        elif num_file<1:
            raise ValueError("no contents in target directory")
        elif (num_file>1) and (identical_ftype==False):
            raise InconsistentError
        elif (num_file>1) and (identical_ftype==True):
            if file_type in ['.csv','.parquet','.pkl']:
                self._load_to_attrDict(contents, single_file=False, file_type= file_type, **kwargs)
            else:
                raise FormatError
        elif num_file==1:
            if file_type in self.supported_format:
                self._load_to_attrDict(contents,single_file=True, file_type = file_type, **kwargs)
            else:
                raise FormatError
        else:
            raise Exception("exception happened during _load_feed_from_source_dir")

        

if __name__ == '__main__':
    pass


import streamlit as st
from .whiteboard import WhiteBoard

class StSupplement:
    _tabs = {
        "whiteboard":True,
        "document": True,
        "file_manager":True
    }


    def __init__(self,label = "supplement",expanded=False, **kwargs):
        self._expander = st.expander(label, expanded=expanded)

        # load **kwargs to adjust which tabs should be put into supplement
        for k,v in self._tabs.items():
            if k in kwargs.keys(): self._tabs[k]=kwargs[k]



    def initialize_expander(self):
        activator = {
            "whiteboard":self.set_whiteboard,
            "document":self.set_document,
            "file_manager":self.set_file_manager
        }

        tab_list = [k for k,v in self._tabs.items() if v is True]
        self._tabs =  self._expander.tabs(tab_list)
        for i,tab in enumerate(tab_list):
            activator[tab](self._tabs[i]) 

        return self._expander

    def set_whiteboard(self,tab):
        with tab:
            WB=WhiteBoard()
    def set_document(self, tab, header:str="Document"):
        with tab:
            st.header(header)
            doc_file = open("./doc/document.md", "r")
            doc_str = doc_file.read()
            st.markdown(doc_str,unsafe_allow_html=True)

    def set_file_manager(self,tab):
        pass





"""
This mapper is for converting the database data type to intended dtype for Pandas Dataframe
"""
import sqlalchemy
import numpy as np

# This dtype_mapper plays as a default setting, but any specific version can be applied for map_db_datatypes_to_dtype by passing the following format as arg datatype_mapper
dtype_mapper = {
    ('oracle', 'cx_oracle'): {
        "<cx_Oracle.DbType DB_TYPE_NUMBER>": np.float32,
        "<cx_Oracle.DbType DB_TYPE_VARCHAR>": 'string',
        "<cx_Oracle.DbType DB_TYPE_TIMESTAMP_LTZ>": 'datetime64[ns]'
        # Add more mappings as needed
    },
    ('mssql', 'pyodbc'): {
        "<class 'int'>": np.int32,
        "<class 'str'>": 'string',
        "<class 'float'>": np.float32,
        # Add more mappings as needed
    }
    # Add mappings for other database types and drivers
}

def get_column_names_and_date_types(result: sqlalchemy.engine.CursorResult):
    column_names = [col[0] for col in result.cursor.description]
    db_data_types = [str(col[1]) for col in result.cursor.description]
    return column_names, db_data_types

def map_db_datatypes_to_dtype(
        database_type:str, 
        database_driver:str, 
        db_data_type:str,
        datatype_mapper:dict=dtype_mapper
        ):
    """map database data type to dtype for Pandas dataframe

    Args:
        database_type (str): engine.dialect.name where engine is Sqlalchemy Engine
        database_driver (str): engine.driver where where engine is Sqlalchemy Engine
        db_data_type (str): str(result.cursor.description[i][1]) for the ith columns of query result where result is the Sqlalchemy CursorResult
        datatype_mapper (dict): map out db data types to dtypes for dataframe

    Returns:
        type/str: dtype for Pandas
    """
    # Retrieve the intended dtype based on the provided database type, driver, and data type; 
    # NOTE: object is the default dtype if the key is not found 
    intended_dtype = datatype_mapper.get((database_type.lower(), database_driver.lower()), {}).get(db_data_type, object)

    return intended_dtype


from typing import List

def str_list_for_IN_condition(str_list:List[str])->str:
    """convert a string list to the format for putting between the bracket () for IN condition while doing SQL query.
    For example, str_list,["A1","A2,"A3"], will be convert to 'A1','A2','A3' and we can inject this result to a predefined sql query like follows:  

            with ARGS AS ( 
                SELECT
                    {str_list} as str_list 
                from DUAL
                )

            SELECT
                column1,
                column2,
            FROM
                table1, ARGS
            WHERE table1.column1 is not NULL
                and table1.column2 in (ARGS.str_list)


    Args:
        str_list (List[str]): a string list

    Returns:
        str: string variable with the format for  IN condition while doing SQL query
    """
    return ','.join(f"'{item}'" for item in str_list) 




def lower_and_replace_space_with_underscore(input_string):
        return input_string.replace(
            '   ', '_'
        ).replace(
            '   ', '_'
        ).replace(
            '  ', '_'
        ).replace(
            ' ', '_'
        ).lower()

def upper_and_replace_space_with_underscore(input_string):
        return input_string.replace(
            '   ', '_'
        ).replace(
            '   ', '_'
        ).replace(
            '  ', '_'
        ).replace(
            ' ', '_'
        ).upper()


from .connector import DBConnector,BlobConnector
import logging
from sqlalchemy.exc import DatabaseError


class DataMigrator:


    def __init__(self, blob_access, db_access):
        """
        Args:
            blob_access (toolbox.dao.connector.db_access): db_access for target blob storage
            db_access (toolbox.dao.connector.db_access): db_access for source database
        """
        self.logger = logging.getLogger(__name__)

        self._db_access = db_access

        self.DBC = DBConnector(db_access)
        self.BC = BlobConnector(blob_access)


    def check_folder(self, target_folder_path:str):
        if target_folder_path not in self.BC.get_blob_list():
            self.BC.add_folder(folder_path=target_folder_path)
            self.logger.info(f'folder `{target_folder_path}` has been created')
            return False
        else:
            self.logger.info('folder `{target_folder_path}` exists and following files are found:\n')
            self.logger.info(self.BC.get_blob_list(target_folder_path))
            return True


    def migrate(
            self,
            period:tuple,
            target_folder_path:str,
            predefined_query_name:str,
            cache_name:str,
            queires_dir: str="./queries",
            chunk_mode:bool=False,
            enforce_dtype:bool=True,
            **pars
    ):
        """migrate data from source db to target Azure Blob Storage

        Args:
            period (tuple(int,)): (start_week,end_week), e.g., (202201,202208)
            target_folder_path (str): path of the target folder
            predefined_query_name (str): name of predefined query for source db
            cache_name (str): label for the cache file(s) without timestamp
            queires_dir (str, optional): where to looking for predefined query. Defaults to "./queries".
            chunk_mode (bool, optional): pull data in chunk mode or not. Defaults to False.
            enforce_dtype (bool, optional): enforce to convert data type for resulted dataframe. Defaults to True.
        
        Example:
        ```python
        from toolbox.dao import DataMigrator
        from toolbox.dao.connector import BlobConnector,DBConnector,parse_db_access
        from toolbox.utility import set_logger

        logger = set_logger('DEBUG')

        blob_access = parse_db_access("db.ini","BLOB_Storage")
        BC = BlobConnector(blob_access)

        db_access = parse_db_access("db.ini","XEUS")
        DBC = DBConnector(db_access)

        DM=DataMigrator(
            blob_access=blob_access,
            db_access=db_access
        )

        DM.migrate(
            period=(202338,202339),
            target_folder_path="test",
            cache_name="test_df",
            queires_dir="./toolbox/dao/queries/",
            predefined_query_name="pull_raw_lot_flow",
            chunk_mode=True,
            enforce_dtype=True
        )
        ```
        """
        
        self.check_folder(target_folder_path)

        for ww in range(period[0],period[1]):

            pars["start_week"] = ww 
            pars["end_week"] = ww+1

            self.DBC.set_queries_dir(queires_dir=queires_dir,**pars)
            self.logger.info(f"Set pars ready for the query in week {ww}")

            try:
                if not chunk_mode:
                    result_df = self.DBC.pull_predefined_query(predefined_query_name,**pars)
                    file_name = f"{pars['start_week']}_{cache_name}.parquet.gzip"
                    self.BC.upload_parquet_from_df(result_df, target_folder_path, file_name, compression='gzip')
                else:
                    dfs_generator = self.DBC.pull_predefined_query(
                            query_name = predefined_query_name,
                            chunk_mode = chunk_mode,
                            enforce_dtype = enforce_dtype,
                            **pars
                    )
                  
                    for i, chunk_df in enumerate(dfs_generator):
                        file_name = f"{ww}_{cache_name}_{i}.parquet.gzip"
                        self.BC.upload_parquet_from_df(chunk_df, target_folder_path, file_name, compression='gzip')
                        self.logger.info(f"Regarding week {ww}: chunk {i} finished")

                self.logger.info(f"The query result of week {ww} is uploaded to Blob Storage")
                
            except DatabaseError as e:
                # Check if the error message contains "ORA-12170"
                if "ORA-12170" in str(e):
                    self.logger.error(f"Task for {ww} got ORA-12170: TNS:Connect timeout occurred")
                else:
                    # Handle other types of DatabaseError
                    self.logger.error(f"Task for {ww} got other DatabaseError: {e}")

        self.logger.info(f"Finish data migration")



from streamlit_drawable_canvas import st_canvas
import streamlit as st

from PIL import Image
import io

class WhiteBoard:
    def __init__(self):
        self.WB = st.container()
        self.pars={}
        with self.WB:

            self.initial_toolbar()
            self.canvas = self.initial_canvas()
            if self.canvas.image_data is not None:
                jpg_col,png_col, empty = st.columns([1,1,8])
                with jpg_col:
                    self.convert_to_jpg()
                with png_col:
                    self.convert_to_png()
                    

    def convert_to_png(self):
        img = Image.fromarray(self.canvas.image_data.astype("uint8"), mode="RGBA")
        buffered = io.BytesIO()
        img.save(buffered, format="PNG")
        img_data = buffered.getvalue()

        return st.download_button(
            label="export png",
            data=img_data,
            file_name='WB.png',
            mime='image/png',
        )
    def convert_to_jpg(self):
        img = Image.fromarray(self.canvas.image_data.astype("uint8"), mode="YCbCr")
        buffered = io.BytesIO()
        img.save(buffered, format="JPEG")
        img_data = buffered.getvalue()

        return st.download_button(
            label="export jpg",
            data=img_data,
            file_name='WB.jpg',
            mime='image/jpg',
        )

    def initial_toolbar(self):
        pars=self.pars

        bg_color_col,bg_image_col, empty = st.columns([1,3,6])
        with bg_color_col:
            pars["bg_color"] = st.color_picker("Background color: ", "#eee")
        with bg_image_col:
            pars["bg_image"] = st.file_uploader("Background image:", type=["png", "jpg"])


        drawing_mode_col,stroke_color_col,stroke_width_col,point_display_radius_col  = st.columns([6,1,2,1])
        with drawing_mode_col:
            pars["drawing_mode"] = st.radio(
                "Mode:",
                ("freedraw", "line", "rect", "circle", "transform", "polygon", "point"),
                horizontal=True
            )

        with stroke_color_col:
            pars["stroke_color"] = st.color_picker("Stroke color: ")

        with stroke_width_col:
            pars["stroke_width"] = st.slider("Stroke width: ", 1, 10, 1)

        with point_display_radius_col:   
            pars["point_display_radius"]= st.slider("Point radius: ", 1, 25, 3)

    def initial_canvas(self):
        pars=self.pars

        return st_canvas(
            fill_color="rgba(255, 165, 0, 0.3)",  # Fixed fill color with some opacity
            stroke_width=pars["stroke_width"],
            stroke_color=pars["stroke_color"],
            background_color=pars["bg_color"],
            background_image=Image.open(pars["bg_image"]) if pars["bg_image"]  else None,
            update_streamlit=True,
            height=900,
            width=1200,
            drawing_mode=pars["drawing_mode"],
            point_display_radius=pars["point_display_radius"] if pars["drawing_mode"] == 'point' else 0,
            key="full_app",
        )



import os
import io
import csv
import logging
import zipfile
import logging

from azure.storage.blob import ContainerClient
import pandas as pd

from ..feed import attrDict
from ...utility import bp

class BlobConnector:
    def __init__(self, db_access):
        self.logger = logging.getLogger(__name__)
        self._db_access = db_access
        self.blob_container_client = self._get_container_client
        

    @property
    def _get_container_client(self,log_only_error=True):
        # Set the logging level for all azure-storage-* libraries

        if log_only_error:
            logger = logging.getLogger('azure.core.pipeline.policies.http_logging_policy').setLevel(logging.ERROR)
        else:
            logger = logging.getLogger('azure-storage').setLevel(logging.ERROR)

        conn_str=''.join(
            [
            f'DefaultEndpointsProtocol={self._db_access["defaultendpointsprotocol"]};',
            f'AccountName={self._db_access["accountname"]};',
            f'AccountKey={self._db_access["accountkey"]};',
            f'EndpointSuffix={self._db_access["endpointsuffix"]}'
            ]
        )

        container_client = ContainerClient.from_connection_string(conn_str=conn_str,
                                                                   container_name=self._db_access["containername"],
                                                                   logger=logger)
        return container_client

    def walk_blob_container(self,*args,**kwargs):
        """returns a generator to list the blobs under the specified container. 
        The generator will lazily follow the continuation tokens returned by the service. 
        This operation will list blobs in accordance with a hierarchy, as delimited by the specified delimiter character.
        Args:
            name_starts_with (str): Filters the results to return only blobs whose names begin with the specified prefix.
            include (list[str] or str): Specifies one or more additional datasets to include in the response. 
                    Options include: 'snapshots', 'metadata', 'uncommittedblobs', 'copy', 'deleted', 'deletedwithversions', 'tags', 'versions', 'immutabilitypolicy', 'legalhold'.
            delimiter (str): When the request includes this parameter, the operation returns a BlobPrefix element in the response body that acts as a placeholder for all blobs whose 
                            names begin with the same substring up to the appearance of the delimiter character. The delimiter may be a single character or a string.
        Returns:
            generator: An iterable (auto-paging) response of BlobProperties.
        Check doc for blob_container_client.walk_blobs to learn more 
        """
        return self.blob_container_client.walk_blobs(args,kwargs)

    def get_blob_list(self,folder_path:str=None)->list[str]:
        """return list of blob names
        If folder_path is None, return all blob names in the container

        Args:
            folder_path (str, optional): path of a specific folder. Defaults to None. If it ends with '/', folder itself won't be included as part of results

        Returns:
            list[str]: list of blob names
        """
        if folder_path is None:
            blob_list = list(self.blob_container_client.list_blob_names())
        else:
            blob_list = [x for x in self.blob_container_client.list_blob_names() if x.startswith(folder_path)]
        return blob_list


    def check_blob_existance(self, blob_name:str)->bool:
        """check a blob exists or not

        Args:
            blob_name (str): path of target blob

        Returns:
            bool: exists or not
        """
        if blob_name.endswith("/"): blob_name=blob_name[:-1]
        return True if blob_name in self.get_blob_list() else False

    @classmethod
    def get_blob_name(cls,file_name:str, blob_path:str)->str:
        """blob_name is the path for accessing a specific blob via blob_container_client
        Conceptually, it will look like <folder_name>/<subfolder_name>/<file_name>
        it only provides file_name as blob_path, it sets container root as like 'working directory'

        Args:
            file_name (str): file name
            blob_path (str): blob path

        Returns:
            blob_name (str): blob_name
        """
        if blob_path is not None:
            if blob_path.startswith("/"): blob_path = blob_path[1:]
            
            if blob_path.endswith("/"):
                blob_name = f"{blob_path}{file_name}"
            else:
                blob_name = f"{blob_path}/{file_name}"
        else:
            blob_name=file_name
        return blob_name

    def del_blob(self,blob_name:str)->None:
        """delete a blob

        Args:
            blob_name (str): path of target blob
        """
        if blob_name.endswith("/"): blob_name=blob_name[:-1]
        if self.check_blob_existance(blob_name):
            container_client = self.blob_container_client
            container_client.delete_blob(blob_name)
        else:
            return None

    def add_folder(self,folder_path:str)->None:
        """Add a empty virtual folder. If the folder exists, all contents will be purged.
        For example, in target container, we already have a virtual folder called <folder_X> and you want to add a empty folder,<folder_Y>, under <folder_X>, set folder_name as '<folder_X>/<folder_Y>'.

        Args:
            folder_path (str): full path of intended virtual folder which should not end with `/`
        """
        if folder_path.endswith('/'):
            self.logger.warning("folder_path for BlobConnector.add_folder() should not end with `/`")
            
            while folder_path.endswith('/'):
                folder_path = folder_path[:-1]

        if self.check_blob_existance(folder_path):
            _ = [self.del_blob(x) for x in self.get_blob_list(folder_path)]
        else:
            dummy_blob = folder_path + "/dummy.txt"
            container_client = self.blob_container_client
            container_client.upload_blob(name=dummy_blob,data=" ")
            container_client.delete_blob(dummy_blob)

    def file_upload(self,file_path:str, blob_path:str=None)-> None:
        """upload local file to Azure Blob Storage; delete blob first, if same blob exists.
           When blob_path is None, file will be put right in the container
           To put file in a given folder, blob_path is required

        Args:
            file_path (str): file path
            blob_path (str): blob path; default in None
        """
        container_client = self.blob_container_client
        file_name = os.path.basename(file_path)
         
        blob_name = self.get_blob_name(file_name,blob_path)

        if self.check_blob_existance(blob_name):
            container_client.delete_blob(blob_name)

        _ = container_client.upload_blob(name=blob_name, data=file_path)

        

    
    def upload_csv_from_df(self, df:pd.DataFrame, blob_path:str, file_name:str, archive:bool=True)->None:
        """convert a DataFrame object to csv then upload to Azure Blob Storage; delete blob first, if same blob exists.
        This method is a stream-base operation. 
        Remember, blob only has virtual folder so files in a folder have blob name including folder name

        Args:
            df (pd.DataFrame): source dataframe
            blob_path (str): path points to the location for uploading the file 
            file_name (str): name of the file
            archive (bool, optional): zip it or not. Defaults to True.

        Raises:
            ValueError: file_name should end by .zip if archive
            ValueError: file_name should end by .csv if not archive
        """

        assert isinstance(file_name,str)
        assert isinstance(blob_path,str)
        blob_name = self.get_blob_name(file_name,blob_path)

        if archive:
            if not file_name.endswith(".zip"): raise ValueError("blob_name should end by .zip") 
        else:
            if not file_name.endswith(".csv"): raise ValueError("blob_name should end by .csv") 

        container_client = self.blob_container_client

        if self.check_blob_existance(blob_name):
            container_client.delete_blob(blob_name)
        
        
        in_memory_csv = io.BytesIO()
        
        if archive:
            df.to_csv(in_memory_csv,
                    compression={"method":'zip',
                    "archive_name":os.path.splitext(file_name)[0]+".csv"}
                    ) #use splitext in stead of basename because of virtual folder in Blob storage
        else:
            df.to_csv(in_memory_csv)

        in_memory_csv.seek(0)
        _ = container_client.upload_blob(name=blob_name, data=in_memory_csv)

    def upload_parquet_from_df(self, df:pd.DataFrame, blob_path:str, file_name:str, compression:str=None)->None:
        """convert a DataFrame object to parquet then upload to Azure Blob Storage; delete blob first, if same blob exists.
        This method is a stream-base operation. 
        Remember, blob only has virtual folder so files in a folder have blob name including folder name
        General Usage : gzip is often a good choice for cold data, which is accessed infrequently. snappy  are a better choice for hot data, which is accessed frequently.

        Args:
            df (pd.DataFrame): source dataframe
            blob_path (str): path points to the location for uploading the file 
            file_name (str): name of the file
            compression (str, optional): 'gzip' or 'snappy'. Defaults to None.

        Raises:
            ValueError: file_name should end by .gzip or .snappy if enable compression
            ValueError: file_name should end by .parquet if no compression
        """

        assert isinstance(file_name,str)
        assert isinstance(blob_path,str)


        match compression:
            case None:
                if not file_name.endswith(".parquet"): raise ValueError("blob_name should end by .parquet") 
            case "gzip":
                if not file_name.endswith(".parquet.gzip"): raise ValueError("blob_name should end by .parquet.gzip") 
            case "snappy":
                if not file_name.endswith(".parquet.snappy"): raise ValueError("blob_name should end by .parquet.snappy") 
            case _:
                raise ValueError("unsupported string is passed as compression") 
            
        blob_name = self.get_blob_name(file_name,blob_path)
        container_client = self.blob_container_client

        if self.check_blob_existance(blob_name):
            container_client.delete_blob(blob_name)
        
        in_memory_parquet = io.BytesIO()
        if compression is not None:
            df.to_parquet(in_memory_parquet,
                    compression={"method":compression}) 
        else:
            df.to_parquet(in_memory_parquet)
        in_memory_parquet.seek(0)
        _ = container_client.upload_blob(name=blob_name, data=in_memory_parquet)
        


    def blob_dump(self,blob_name:str, dir_path: str = ".", to_memory:bool=False, to_dataframe:bool=False,to_attrDict:bool=False):
        """download specified blob to the specified location

        Args:
            blob_name (str): blob_path+file_name, path of the blob because Blob storage uses virtual folder
            dir_path (str, optional): specified location of keeping the blob file. Defaults to ".".
            to_memory (bool, optional): assign in-momery zip object as return type if True. Defaults to False.
            to_dataframe (bool, optional): assign pandas dataframe as return type. It works only if the blob is csv/xlsx/parquet. Defaults to False.
            to_attrDict (bool, optional): assign toolbox.dao.feed.attrDict as return type if True. It works only if the blob is an archive of csv file(s). Defaults to False.

        Raises:
            ValueError: if target container doesn't have any blob named blob_name

        Returns:
            _type_: Defaults to None; But toolbox.dao.feed.attrDict if to_attrDict else in-momery zip object if to_memory 
        """

        if not self.check_blob_existance(blob_name):
            self.logger.info(f"target container doesn't have any blob named {blob_name}")
            raise ValueError(f"target container doesn't have any blob named {blob_name}")

        if not os.path.exists(dir_path): 
            os.makedirs(dir_path)
            self.logger.info(f"create required folder(s) because of non-exist path {dir_path}")
            
        blob_client=self.blob_container_client

        file_path = os.path.join(dir_path, os.path.basename(blob_name)) 
        download_stream = blob_client.download_blob(blob=blob_name)

        # Just download the original file to local
        if not (to_attrDict or to_memory or to_dataframe):
            with open(file_path, "wb") as target_blob:
                target_blob.write(download_stream.readall())
            return None

        # download zip file
        elif blob_name.endswith('.zip'):
            in_memory_zip = io.BytesIO(download_stream.readall())
            zf = zipfile.ZipFile(in_memory_zip)

            # further convert to dataframe
            if to_dataframe:
                if len(list(zf.namelist()))>1:
                    self.logger.error("only work with single file when to_dataframe=True")
                    raise Exception("only work with single file when to_dataframe=True")
                
                filename=list(zf.namelist())[0]
                file_ext = os.path.splitext(filename)[1]
                to_df = {
                    '.xlsx':pd.read_excel,
                    '.csv': pd.read_csv
                }

                try:
                    df = to_df[file_ext](zf.open(filename))
                except:
                    self.logger.error(f"Can't convert {filename} in zip blob {blob_name} to a dataframe")
                    raise Exception(f"Can't convert {filename} in zip blob {blob_name} to a dataframe")
                finally:
                    return df

            elif to_attrDict:
                
                dict_DF = attrDict(
                    {
                        os.path.splitext(os.path.basename(name))[0]: pd.read_csv(io.StringIO(str(zf.read(name),'utf-8'))) for name in zf.namelist() if name.endswith(".csv") 
                    }
                )

                return dict_DF
            
            elif to_memory:
                return zf


        elif blob_name.endswith(('.parquet.gzip','.parquet.snappy','.parquet')):
            if to_dataframe:
                try:
                    stream = io.BytesIO(download_stream.readall())
                    stream.seek(0)
                    return pd.read_parquet(stream)
                except Exception as e:
                
                    self.logger.error(e)
                    raise Exception(f"Fail to convert parquet blob {blob_name} to a dataframe")
            else:
                raise Exception("Only support directly dumping or converting to dataframe for parquet")

        else:
            if not (blob_name.endswith('.csv') or blob_name.endswith('.xlsx')):
                self.logger.error("Except an archive of csv/xlsx, it only accepts a single csv or xlsx")
                raise Exception("Except an archive of csv/xlsx, it only accepts a single csv or xlsx")

            elif (to_dataframe or to_attrDict):

                file_ext = os.path.splitext(blob_name)[1]
                to_df = {
                    '.xlsx':pd.read_excel,
                    '.csv': pd.read_csv
                }

                if file_ext == '.xlsx':
                    blob =download_stream.content_as_bytes()
                else:
                    stream = io.BytesIO()
                    download_stream.download_to_stream(stream)
                    stream.seek(0)
                    blob = stream
                
                try:
                    df = to_df[file_ext](blob)
                except:
                    self.logger.error(f"Can't convert spreadsheet blob {blob_name} to a dataframe")
                    raise Exception(f"Can't convert spreadsheet blob {blob_name} to a dataframe")

                return df

            elif to_memory:
                return download_stream






from .from_file_loader import fl_to_csv,fl_to_excel,fl_to_json,fl_to_sqlite


MAPPER = {
    "toolbox.dao.files.loader.FileLoader": {
        "to_json":fl_to_json,
        "to_csv":fl_to_csv,
        "to_sqlite":fl_to_sqlite,
        "to_excel":fl_to_excel,
    },
}


from collections import namedtuple

def namedtuple_from_dict(source_dict:dict,name:str)->namedtuple:
    """Create and return a namedtuple instance with fields populated from a dictionary.
    This function dynamically creates a new namedtuple class with a given name and fields corresponding to the keys of the source dictionary. The values of the namedtuple are filled with the values from the source dictionary.

    Args:
        source_dict (dict): The dictionary whose keys and values are to be used to create the namedtuple.
        name (str): The name of the new namedtuple class to be created.

    Returns:
        namedtuple: An instance of the newly created namedtuple class, populated with data from the source_dict.
    """

    return namedtuple(name,source_dict.keys())(**source_dict)


import pathlib
import os
from collections.abc import KeysView
import pandas as pd


def all_equal(iterator):
    iterator = iter(iterator)
    try:
        first = next(iterator)
    except StopIteration:
        return True
    return all(first == x for x in iterator)

    
def count_folder_number(contents:list) -> int:
    #return sum(os.path.isdir(c) for c in contents)
    return sum(pathlib.Path(c).is_dir() for c in contents)

def count_file_number(contents:list) -> int:
    return sum(pathlib.Path(c).is_file() for c in contents)

def create_nonduplicated_dict_key(dict_key:str,keys:KeysView, addition = "_") -> str:
    """If dict_key has not existed in keys, directly return, otherwise, keep adding additional string as suffix until no duplication is confirmed

    Args:
        dict_key (str): dictionary key that plans to add
        keys (KeysView): keys of target dictionary
        addition (str, optional): aditional string as suffix to avoid duplication. Defaults to "_".

    Raises:
        TypeError: keys needs to be a dict_keys

    Returns:
        str: confirmed key for futher insert 
    """
    if not isinstance(dict_key.keys(),KeysView):
        raise TypeError('keys needs to be a dict_keys')

    if dict_key not in list(keys):
        return dict_key
    else:
        return create_nonduplicated_dict_key(''.join((dict_key,addition)),keys)

def ensure_directory_exists(output_path: str, create_if_nonexistence:bool=True):
    """
    Check if the directory for the output path exists, and create it if it doesn't.

    Args:
        output_path (str): The path where the output will be saved.
    """
    directory = os.path.dirname(output_path)

    # handle the case if only file name provided as output path, set it as current working directory
    if directory=='': 
        directory='.'

    # Check if the directory exists, and create it if it doesn't
    if not os.path.exists(directory):
        if create_if_nonexistence:
            os.makedirs(directory, exist_ok=True)
        return False
    else:
        return True
    

def df_dq_check(df, spec):
    """
    Perform data quality checks on a DataFrame based on the provided specification.

    Parameters:
    df (pd.DataFrame): The DataFrame to validate.
    spec (pd.DataFrame): A DataFrame containing the specification for validation. 
        It must have the following columns:
        - column_name (str): The expected column names in the DataFrame.
        - data_type (str): The expected data type for each column. Supported types are:
            - 'str': String type.
            - 'int': Integer type.
            - 'float': Float type.
            - 'datetime64[ns]': Datetime type.
        - re_pattern (str, optional): A regular expression pattern to validate string columns. 
          This is only applicable for columns with 'str' data_type.

    Raises:
    ValueError: If the number of columns in the DataFrame does not match the specification.
    ValueError: If the column names in the DataFrame do not match the expected column names.

    Returns:
    pd.DataFrame: The validated and transformed DataFrame. If any rows are invalid, an 
    'Invalid_Row_Message' column is added to indicate the issues.
    # Example: Creating a data specification for a CSV file using JSON

    ## Step 1: Define the data specification in JSON format
    data_spec_json = {
        "Corridor_LT": [
            {
                "column_name": "Corridor",
                "data_type": "str",
                "re_pattern": "LIAM\\s*[A-Za-z0-9]+-\\d+$"
            },
            {
                "column_name": "Location",
                "data_type": "str",
                "re_pattern": "^[A-Z0-9]{3}$"
            },
            {
                "column_name": "LT",
                "data_type": "int",
                "re_pattern": None
            }
        ]
    }

    ## Step 2: Save the JSON to a file
    with open('data_spec.json', 'w') as json_file:
        json.dump(data_spec_json, json_file, indent=4)

    ## Step 3: Load the JSON file and extract the required specification
    with open('data_spec.json', 'r') as json_file:
        loaded_spec = json.load(json_file)

    ## Extract the specification for the desired CSV file
    spec = pd.DataFrame(loaded_spec["Corridor_LT"])
    print(spec)
    """
    # Check if the number of columns matches the expected number
    if len(df.columns) < len(spec.column_name):
        raise ValueError(f"Expected {len(spec.column_name)} columns for the input file or dataframe, which are: {spec.column_name}")

    # assign column names based on their order
    df = df.iloc[:, :spec.column_name.size]
    df.columns = spec.column_name

    # Add a column to indicate if the row is invalid
    df['Invalid_Row_Message'] = ""



    # Validate and transform columns based on the configuration
    for row in spec.itertuples():
        dtype = row.data_type
        re_pattern = row.re_pattern
        if dtype == 'datetime64[ns]':
            df[row.column_name] = pd.to_datetime(df[row.column_name], errors='coerce')
            invalid_rows = df[row.column_name].isnull()
            if invalid_rows.any():
                error_msg_list = [f"Invalid date format in column '{row.column_name}'"] * sum(invalid_rows)
                existing_msg_list = list(df.loc[invalid_rows, 'Invalid_Row_Message'])
                df.loc[invalid_rows, 'Invalid_Row_Message'] = list(map(lambda x:''.join([x[0],";",x[1]]), zip(existing_msg_list,error_msg_list)))
        elif dtype == 'str' and re_pattern:
            invalid_rows = ~df[row.column_name].str.match(re_pattern)
            if invalid_rows.any():
                error_msg_list = [f"Values do not match the expected pattern in column '{row.column_name}'"] * sum(invalid_rows)
                existing_msg_list = list(df.loc[invalid_rows, 'Invalid_Row_Message'])
                df.loc[invalid_rows, 'Invalid_Row_Message'] = list(map(lambda x:''.join([x[0],";",x[1]]), zip(existing_msg_list,error_msg_list)))
        elif dtype in ('int','float'):
            invalid_rows = pd.to_numeric(df[row.column_name], errors='coerce').isnull()
            if invalid_rows.any():
                error_msg_list = [f"Column '{row.column_name}' could not be converted to {dtype}"] * sum(invalid_rows)
                existing_msg_list = list(df.loc[invalid_rows, 'Invalid_Row_Message'])
                df.loc[invalid_rows, 'Invalid_Row_Message'] = list(map(lambda x:''.join([x[0],";",x[1]]), zip(existing_msg_list,error_msg_list)))
        
        if sum(invalid_rows) == 0:
            df[row.column_name] = df[row.column_name].astype(dtype)


    # Drop the Invalid_Row_Message column if it is empty for all rows
    if df['Invalid_Row_Message'].isna().all():
        df.drop(columns=['Invalid_Row_Message'], inplace=True)
    else:
        # Remove trailing semicolon and space from Invalid_Row_Message if any
        df['Invalid_Row_Message'] = df['Invalid_Row_Message'].str.lstrip('; ')

    return df


# Steps for updating this package
# 1. check files includes private/security info and make sure set .gitignore correctly
# 2. update markdown document associated to changes, index.md within python modules
# 3. update README.md, toolbox/change_log.md
# 4. update setup.py, set a new version number, check if any python package is required by changes and add into the file content
# 5. delete old archived files under <toolbox-project-directory>/dist/
# 6. create new archive files for this ditrtibution
#   #install or upgrade the setuptools and wheel packages in the user's home directory
#   python3 -m pip install --user --upgrade setuptools wheel 
#   #create archive files (could be found in folder dist/)
#   python3 setup.py sdist bdist_wheel
# 7. upload the archives to PyPi (login to Pypi and generate API tocken if not on hand)
#   #Check that pacakge twine has been installed
#   python3 -m pip install --user --upgrade twine 
#   twine upload dist/* 
# 8. test new distribution via pip install liamhsieh-toolbox
# 9. push all changes to Github

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="liamhsieh_toolbox",
    version = '0.3.6.7',
    description = "Collections of Python utility",
    author = 'Liam Y. Hsieh, PhD',
    author_email = 'liamhsieh@ieee.org',
    long_description = long_description,
    long_description_content_type = "text/markdown",
    project_urls={
        'Homepage': 'https://github.com/liam-hsieh/liamhsieh-toolbox',
        'Docs':'https://liam-hsieh.github.io/liamhsieh-toolbox/'
    },
    packages = setuptools.find_packages(),
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 2 - Pre-Alpha",
    ],
    #packages = ['toolbox'], #, since I only have a subfolder, toolbox, below setup.py and I don't wanna include docs as well, I just manually list what I need  
    python_requires='>=3.10',
    install_requires=[
        'pandas>=2.2',
        'sqlalchemy>=2.0',
        'openpyxl>=3.0.10',
        'xlsxwriter>=3.2.0',
        'pyxlsb>=1.0.9',
        'dask>=2024.5.0',
        'filetype>=1.1.0',
        'cx-Oracle>=8.3.0',
        'pyodbc>=4.0.0',
        'scipy>=1.9.0',
        'scikit-learn>=1.1.2',
        'matplotlib>=3.6.0',
        'pympler>=1.0.0',
        'pyarrow>=9.0.0',
        'streamlit>=1.13.0',
        'streamlit-drawable-canvas>=0.9.2',
        'streamlit-aggrid>=0.3.3',
        'azure-storage-blob>=12.14.0',
        'pymysql>=1.1',
        'pipe',
        'snowflake-sqlalchemy'
    ]
)



from functools import partial

class BoundFuncAsClsMethod:
    def __init__(self, mapper: dict, class_to_attach_on = None, simple_mode = True):
        """_summary_

        Args:
            mapper (dict): _description_
            class_to_attach_on (_type_, optional): _description_. Defaults to None.
            simple_mode (bool, optional): _description_. Defaults to True.
        Example:
            ################ data required to create an instance of BoundFuncAsClsMethod ##############
            def m1(self):
                return f"{self.user_name}'s method 1"

            def m2(self):
                return self.msg

            def m3(self,favorite):
                return f"{self.user_name} likes {favorite}"

            method_mapper = {
                "method1":m1,
                "method2":m2,
                "method3":m3,
            }
            ############################################################################################

            
            # simple mode
            class Tester:
                def __init__(self,user_name):
                    self.user_name=user_name
                    self.methods = BoundFuncAsClsMethod(method_mapper)

            tester=Tester("Liam")
            tester.msg="original msg"

            # method 1 can access tester.user_name
            print(tester.methods.method1(tester))

            # method2 can access tester.msg while instance of Tester has been passed to m2 as an arg during initilization
            print(tester.methods.method2(tester))

            # modify the content of msg
            tester.msg="new one"

            # the output also changes
            print(tester.methods.method2(tester))

            # testing how it handles additional argument 
            print(tester.methods.method3(tester,"guava"))

            # Non simple mode
            class Tester:
                def __init__(self,user_name):
                    self.user_name=user_name
                    self.methods = BoundFuncAsClsMethod(method_mapper, self, simple_mode=False)

            tester=Tester("Liam")
            tester.msg="original msg"

            # passing or not passing tester as arg for mehtod1 has the same result
            print(tester.methods.method1())

            # passing or not passing tester as arg for mehtod1 has the same result
            print(tester.methods.method2())

            # modify the content of msg
            tester.msg="new one"

            # the output also changes
            print(tester.methods.method2())

            # testing how it handles additional argument 
            print(tester.methods.method3("guava"))

            # Output
            In this example, both modes will have the same output as follows:
            Liam's method 1
            original msg
            new one
            Liam likes guava
        """
        if simple_mode:
            # read through method_mapper and iteratively assign each value to a new property using key value
            for method_name, method in mapper.items():
                setattr(self,method_name, method)
        else:
            assert class_to_attach_on, "When using non simple mode, simple_mode=False, class_to_attch_on needs to be passed for initilizing the instance of BoundFuncAsClsMethod"
            
            for method_name, method in mapper.items():
                # bound_method = lambda self=class_to_attach_on, method=method: method(self)
                # setattr(self, method_name, bound_method)
                # Bind the method to class_to_attach_on and allow additional arguments to be passed later
                bound_method = partial(method, class_to_attach_on) 
                setattr(self, method_name, bound_method)





import datetime as dt
from typing import (
    Callable,
)

def timeit(method: Callable) -> Callable:
    def time_and_run(obj, *args, **kw):
        ts = dt.datetime.now()
        result = method(obj, *args, **kw)
        te = dt.datetime.now()

        query_name = None
        if method.__name__ == 'pull_predefined_query':
            query_name = args[0]

        record = {
            'code': query_name or method.__name__,
            'start_time': ts,
            'end_time': te,
            'run time(sec)': (te - ts).total_seconds(),
        }
        if hasattr(obj,"watcher"):
            obj.watcher["run_time"] = record
        else:
            watcher = {}
            setattr(obj,"watcher",watcher)
            obj.watcher["run_time"] = record
        return result

    return time_and_run